@@ -8,6 +8,7 @@ from dateutil.relativedelta import relativedelta
 import json
 
 from colorama import Style, Fore
+from builtins import TimeoutError, ConnectionResetError
 
 DEFAULT_HISTORIC_DATA_ID = 1000
 DEFAULT_MARKET_DATA_ID = 2000
@@ -84,6 +85,7 @@ class LiveClient(WebSocketApp):
                 self.parent_app.touchline_data[ticker_id].oi = int(contract_details[8])
                 self.parent_app.touchline_data[ticker_id].prev_oi = int(contract_details[9])
                 self.parent_app.touchline_data[ticker_id].turnover = float(contract_details[10])
+                self.parent_app.touchline_data[ticker_id].tick_type = 0
                 self.parent_app.live_data[ticker_id].populate_using_touchline(self.parent_app.touchline_data[ticker_id])
 
     def handle_trade_data(self, trade_tick):
@@ -106,6 +108,7 @@ class LiveClient(WebSocketApp):
                     elif special_tag == 'L':
                         self.parent_app.live_data[ticker_id].day_low = self.parent_app.touchline_data[ticker_id].low = ltp
                 self.parent_app.live_data[ticker_id].tick_seq = int(trade_tick[8])
+                self.parent_app.touchline_data[ticker_id].tick_type = 1
                 # Calculating addn data
                 self.parent_app.live_data[ticker_id].calculate_additional_data()
                 try:
@@ -133,6 +136,7 @@ class LiveClient(WebSocketApp):
                 self.parent_app.live_data[ticker_id].best_bid_qty = int(bidask_tick[3])
                 self.parent_app.live_data[ticker_id].best_ask_price = float(bidask_tick[4])
                 self.parent_app.live_data[ticker_id].best_ask_qty = int(bidask_tick[5])
+                self.parent_app.touchline_data[ticker_id].tick_type = 2
         except KeyError:
             pass
         except Exception as e:
@@ -164,11 +168,13 @@ class LiveClient(WebSocketApp):
             print(f'{Style.BRIGHT}{Fore.RED}Encountered with bar feed - {e}{Style.RESET_ALL}')
 
     def ws_error(self, error):
-        print(error)
+        if any(isinstance(error, conn_error) for conn_error in [ConnectionResetError, TimeoutError]):
+            print('Alright...')
+            print(error)
 
-    def ws_open(self, message):
-        # self.sock.settimeout(300)
-        print(message)
+    def ws_open(self):
+        print(f'SSL?? {self.sock.is_ssl()}')
+        self.sock.settimeout(300)
 
 
 class HistoricalWebsocket:
@@ -190,39 +196,55 @@ class HistoricalWebsocket:
             print(f"{Style.BRIGHT}{Fore.RED}Failed to connect with error message = {welcome_msg['message']}{Style.RESET_ALL}")
             self.hist_socket.close()
 
+    def get_historic_data(self, contract, end_time, start_time, bar_size):
+        if bar_size.lower() == 'tick':
+            return self.get_hist_tick_data(contract, end_time, start_time)
+        elif bar_size.lower() == 'eod':
+            return self.get_hist_eod_data(contract, end_time, start_time)
+        else:
+            bar_size = bar_size.replace(' ', '')
+            if bar_size[-1] == 's':
+                bar_size = bar_size[:-1]
+            return self.get_hist_bar_data(contract, end_time, start_time, bar_size)
+
     def get_hist_bar_data(self, contract, end_time, start_time, bar_size):
         self.hist_socket.send(f'{{"method": "gethistory", "interval": "{bar_size}", "symbol": "{contract}", "from": "{start_time}", "to": "{end_time}"}}')
         raw_hist_data = self.hist_socket.recv()
         json_response = json.loads(raw_hist_data)
         if json_response['success']:
             hist_data = json_response['data']
-            hist_data = self.hist_bar_data_to_dict_list(hist_data, contract)
+            hist_data = self.hist_bar_data_to_dict_list(hist_data)
+            return hist_data
+        else:
+            print(f'{Style.BRIGHT}{Fore.RED}ERROR: Your request failed with error "{json_response["message"]}"{Style.RESET_ALL}')
+            return []
+
+    def get_n_historic_bars(self, contract, end_time, no_of_bars, bar_size):
+        self.hist_socket.send(f'{{"method": "gethistorylastnbars", "interval": "{bar_size}", "symbol": "{contract}", "nbars": "{no_of_bars}", "to": "{end_time}"}}')
+        raw_hist_data = self.hist_socket.recv()
+        json_response = json.loads(raw_hist_data)
+        if json_response['success']:
+            hist_data = json_response['data']
+            hist_data = self.hist_bar_data_to_dict_list(hist_data)
             return hist_data
         else:
             print(f'{Style.BRIGHT}{Fore.RED}ERROR: Your request failed with error "{json_response["message"]}"{Style.RESET_ALL}')
             return []
 
     @staticmethod
-    def hist_bar_data_to_dict_list(hist_data, contract):  # No need for symbol other than printing
+    def hist_bar_data_to_dict_list(hist_data):
         data_list = []
-        count = 0
         for j in hist_data:
-            try:  # TODO: Remove HOT-FIX because of this data point ['2020-02-13T09:15:00', 31565.05, 31565.05, 31565.05, 31565.05, 160, None]
-                data_list.append({'time': datetime.strptime(j[0], '%Y-%m-%dT%H:%M:%S'),
-                                  'o': float(j[1]),
-                                  'h': float(j[2]),
-                                  'l': float(j[3]),
-                                  'c': float(j[4]),
-                                  'v': int(j[5]),
-                                  'oi': int(j[6])})
-            except TypeError:
-                print(f'{Style.BRIGHT}{Fore.RED} {contract} erred with {j}...{Style.RESET_ALL} \n\t {hist_data[count-2]} \n\t {hist_data[count-1]} \n\t {hist_data[count]} \n\t {hist_data[count+1]} \n\t {hist_data[count+2]}')
-                continue
-            count = count + 1
+            data_list.append({'time': datetime.strptime(j[0], '%Y-%m-%dT%H:%M:%S'),
+                              'o': float(j[1]),
+                              'h': float(j[2]),
+                              'l': float(j[3]),
+                              'c': float(j[4]),
+                              'v': int(j[5]),
+                              'oi': int(j[6])})
         return data_list
 
     def get_hist_tick_data(self, contract, end_time, start_time):
-        print(f'{{"method": "gethistory", "interval": "tick", "symbol": "{contract}", "from": "{start_time}", "to": "{end_time}"}}')
         self.hist_socket.send(f'{{"method": "gethistory", "interval": "tick", "symbol": "{contract}", "from": "{start_time}", "to": "{end_time}"}}')
         raw_hist_data = self.hist_socket.recv()
         json_response = json.loads(raw_hist_data)
@@ -239,8 +261,8 @@ class HistoricalWebsocket:
         data_list = []
         count = 0
         for j in hist_data:
-            try:  # TODO: Remove HOT-FIX because of any erroneous data points
-                try:
+            try:
+                try:  # TODO: Remove HOT-FIX because of any erroneous data points
                     oi = int(j[3])
                 except TypeError:  # If OI is None
                     print(f'{Style.BRIGHT}{Fore.RED} {contract} erred with {j}...{Style.RESET_ALL} \n\t {hist_data[count - 2]} \n\t {hist_data[count - 1]} \n\t {hist_data[count]} \n\t {hist_data[count + 1]} \n\t {hist_data[count + 2]}')
@@ -264,6 +286,33 @@ class HistoricalWebsocket:
             count = count + 1
         return data_list
 
+    def get_hist_eod_data(self, contract, end_time, start_time):
+        start_time = start_time.split('T')[0]
+        end_time = end_time.split('T')[0]
+        self.hist_socket.send(f'{{"method": "gethistory", "interval": "EOD", "symbol": "{contract}", "from": "{start_time}", "to": "{end_time}"}}')
+        raw_hist_data = self.hist_socket.recv()
+        json_response = json.loads(raw_hist_data)
+        if json_response['success']:
+            hist_data = json_response['data']
+            hist_data = self.hist_eod_data_to_dict_list(hist_data, contract)
+            return hist_data
+        else:
+            print(f'{Style.BRIGHT}{Fore.RED}ERROR: Your request failed with error "{json_response["message"]}"{Style.RESET_ALL}')
+            return []
+
+    @staticmethod
+    def hist_eod_data_to_dict_list(hist_data, contract):  # No need for symbol other than printing
+        data_list = []
+        for j in hist_data:
+            data_list.append({'time': datetime.strptime(j[0], '%d-%m-%Y'),
+                              'o': float(j[1]),
+                              'h': float(j[2]),
+                              'l': float(j[3]),
+                              'c': float(j[4]),
+                              'v': int(j[5]),
+                              'oi': int(j[6])})
+        return data_list
+
 
 class TD:
     def __init__(self, login_id, password, broker_token=None, url='push.truedata.in', live_port=8082, historical_port=8092, *args):
@@ -354,6 +403,19 @@ class TD:
                                                             start_time=start_time,
                                                             bar_size=bar_size)
 
+    def get_n_historical_bars(self, contract,
+                              ticker_id=DEFAULT_HISTORIC_DATA_ID,
+                              end_time=None,
+                              no_of_bars: int = 1,
+                              bar_size="1 min"):
+        global DEFAULT_HISTORIC_DATA_ID
+        end_time = end_time.strftime('%Y-%m-%d') + 'T23:59:59'
+
+        hist_data = self.historical_websocket.get_n_historic_bars(contract, end_time, no_of_bars, bar_size)
+        DEFAULT_HISTORIC_DATA_ID = DEFAULT_HISTORIC_DATA_ID + 1
+        self.hist_data[ticker_id] = hist_data
+        return hist_data
+
     def get_historical_data_from_duration(self, contract,
                                           ticker_id=DEFAULT_HISTORIC_DATA_ID,
                                           end_time=None,
@@ -370,13 +432,7 @@ class TD:
         end_time = end_time.strftime('%Y-%m-%d') + 'T23:59:59'
         start_time = start_time.strftime('%Y-%m-%d') + 'T00:00:00'
 
-        if bar_size == 'tick':
-            hist_data = self.historical_websocket.get_hist_tick_data(contract, end_time, start_time)
-        else:
-            bar_size = bar_size.replace(' ', '')
-            if bar_size[-1] == 's':
-                bar_size = bar_size[:-1]
-            hist_data = self.historical_websocket.get_hist_bar_data(contract, end_time, start_time, bar_size)
+        hist_data = self.historical_websocket.get_historic_data(contract, end_time, start_time, bar_size)
         DEFAULT_HISTORIC_DATA_ID = DEFAULT_HISTORIC_DATA_ID + 1
         self.hist_data[ticker_id] = hist_data
         return hist_data
@@ -392,19 +448,15 @@ class TD:
             end_time = end_time.strftime('%Y-%m-%d') + 'T23:59:59'
         else:
             assert type(end_time) == datetime
+            end_time = end_time.strftime('%Y-%m-%dT%H:%M:%S')
         if start_time is None:
             start_time = datetime.today()
             start_time = start_time.strftime('%Y-%m-%d') + 'T00:00:00'
         else:
             assert type(start_time) == datetime
+            start_time = start_time.strftime('%Y-%m-%dT%H:%M:%S')
 
-        if bar_size == 'tick':
-            hist_data = self.historical_websocket.get_hist_tick_data(contract, end_time, start_time)
-        else:
-            bar_size = bar_size.replace(' ', '')
-            if bar_size[-1] == 's':
-                bar_size = bar_size[:-1]
-            hist_data = self.historical_websocket.get_hist_bar_data(contract, end_time, start_time, bar_size)
+        hist_data = self.historical_websocket.get_historic_data(contract, end_time, start_time, bar_size)
         DEFAULT_HISTORIC_DATA_ID = DEFAULT_HISTORIC_DATA_ID + 1
         self.hist_data[ticker_id] = hist_data
         return hist_data
